from .. import Handler as Base
from .message import Ping, Pong

from bidict import bidict


class Handler(Base):
    VERSION = 1

    BYTE_ORDER = 'big'
    ID_SIZE = 1 # bytes

    MSG_BY_ID = bidict({
        1: Ping,
        2: Pong,
    })

    def server(self, master):
        while msg := self.recv():
            print(f'     {self.addr} -> {msg}')
            reply = None

            if isinstance(msg, Ping):
                reply = Pong(msg.nonce)

            else:
                raise ValueError(f'unhandled message: {msg}')

            if reply:
                print(f'     {self.addr} <- {reply}')
                self.send(reply)

    def send(self, msg):
        self.sock.sendall(self.get_id_bytes(msg))
        msg.send(self)

    def recv(self):
        id_bytes = self.recv_bytes(self.ID_SIZE)
        if id_bytes is None: return None
        msg = self.get_msg_type(id_bytes)
        return msg.recv(self)

    def get_id_bytes(self, msg):
        try:
            mid = self.MSG_BY_ID.inverse[type(msg)]
            try:
                return mid.to_bytes(self.ID_SIZE, self.BYTE_ORDER)

            except OverflowError as e:
                raise ValueError(f'msg type id too big: {mid}') from e

        except KeyError as e:
            raise ValueError(f'unsupported message type: {type(msg)}') from e

    def get_msg_type(self, id_bytes):
        mid = int.from_bytes(id_bytes, self.BYTE_ORDER)
        try:
            return self.MSG_BY_ID[mid]

        except KeyError as e:
            raise ValueError(f'unsupported message type id: {mid}') from e




from .protocol import get_handler, HELLO, KEEPALIVE, VERSION
import lank.node.db as ldb

from gevent import socket, wait #, spawn
from gevent.pool import Pool
from gevent.server import StreamServer
from bidict import bidict
from ntplib import NTPClient

from uuid import UUID
from datetime import datetime, timedelta, timezone


DEFAULT_PORT = 42024

HELLO_SIZE = len(HELLO)
HELLO_TIMEOUT = 9 # seconds

GENERAL_TIMEOUT = KEEPALIVE * 2 # seconds

NODES = [ # FIXME -- this is temporary (put in database?)
    ('localhost', 42024),
    ('localhost', 42124),
    ('72.202.195.53', 42024),
    ('ruckusist.com', 42024),
]

NODES_MIN = 3
NODES_MAX = 9
NODES_WAIT = 3 * 60 # seconds

NTP = 'pool.ntp.org'


class Master:
    def __init__(self, port=DEFAULT_PORT):
        self.port = port

        uuid = ldb.get_meta(ldb.META_NODE_UUID)
        assert uuid
        self.uuid = UUID(uuid)
        print(f'   our uuid is {self.uuid}')

        print(f'   getting time from {NTP}...')
        ntp = NTPClient().request(NTP, version=3)
        self.offset = ntp.offset
        print(f'      our clock is {abs(self.offset)} seconds ', end='')
        if self.offset < 0: print('fast')
        else: print('slow')

        self.pool = Pool()
        self.stream_server = StreamServer(('0.0.0.0', self.port), self.server,
            spawn=self.pool)

        self.buffer = bytearray(HELLO_SIZE)

        #self.peers_by_sock = bidict({ })
        #self.labels = { }

        self.labels_by_id = bidict({ })

        self.nodes_by_uuid = { }
        self.nodes_client = { }

        self.reservations = { }
        self.registrations = { }

        self.signed_recently = { }

    def run(self):
        for label in ldb.list_labels():
            self.labels_by_id[label['id']] = label['name']

        print(f'S  listening on port {self.port}')
        self.stream_server.start()

        while True:
            if len(self.nodes_by_uuid) < NODES_MIN:
                # FIXME: limit spawning to NODES_MAX - len(self.nodes)
                #        and then wait

                for addr in NODES:
                    if len(self.nodes_by_uuid) >= NODES_MAX:
                        break
                    if addr in self.nodes_client:
                        continue

                    self.pool.spawn(self.client, addr)

            wait(timeout=NODES_WAIT)
            self.status()

    def now(self):
        return datetime.now(timezone.utc) + timedelta(seconds=self.offset)

    def status(self):
        nodes = len(self.nodes_by_uuid)
        time = self.now().isoformat()
        print(f'** STATUS ** nodes={nodes} ** time={time}')

    def broadcast_nodes(self, msg, skip=None):
        print(f'B    (NODES) <- {msg}')
        self.pool.spawn(self._broadcast_nodes_, msg, skip)

    def _broadcast_nodes_(self, msg, skip=None):
        for handler in self.nodes_by_uuid.values():
            if skip == handler:
                continue

            handler.send(msg)

    def client(self, addr):
        print(f'C+ connecting to {addr}')
        self.nodes_client[addr] = True

        try:
            sock = socket.create_connection(addr, timeout=HELLO_TIMEOUT)
            sock.settimeout(HELLO_TIMEOUT)

            try:
                handler = get_handler(sock, addr, VERSION)
                handler.hello()
                sock.settimeout(GENERAL_TIMEOUT)

                try:
                    handler.client(self)
                    print(f'C- finished {addr}')

                except KeyError as e:
                    print(f'C- terminated {addr} [DENIED: {e}]')
                    if 'NodeIsSelf' in str(e):
                        self.nodes_client[addr] = False

                except ValueError as e:
                    print(f'C- terminated {addr} [BAD MESSAGE: {e}]')

                except socket.timeout:
                    print(f'C- terminated {addr} [GENERAL TIMEOUT]')

            except BrokenPipeError:
                print(f'C- closed {addr} [BROKEN PIPE]')

            except ConnectionResetError:
                print(f'C- closed {addr} [CONNECTION RESET]')

            except socket.timeout:
                print(f'C- terminated {addr} [HELLO TIMEOUT]')

        except ConnectionRefusedError:
            print(f'C- closed {addr} [CONNECTION REFUSED]')

        except socket.timeout:
            print(f'C- terminated {addr} [CONNECT TIMEOUT]')

        finally:
            if self.nodes_client[addr]:
                del self.nodes_client[addr]

    def server(self, sock, addr):
        print(f'S+ connection from {addr}')
        sock.settimeout(HELLO_TIMEOUT)

        try:
            read = sock.recv_into(self.buffer)

            if read == HELLO_SIZE:
                if self.buffer == HELLO:
                    sock.settimeout(GENERAL_TIMEOUT)

                    try:
                        protocol = get_handler(sock, addr)

                        if protocol:
                            try:
                                protocol.server(self)
                                print(f'S- finished {addr}')

                            except ValueError as e:
                                print(f'S- terminated {addr}' \
                                    + f' [BAD MESSAGE: {e}]')

                        else:
                            print(f'S- closed {addr} [CLIENT ABORT]')

                    except ValueError as e:
                        print(f'S- terminated {addr} [PROTOCOL VERSION]')

                    except socket.timeout:
                        print(f'S- terminated {addr} [GENERAL TIMEOUT]')

                else:
                    print(f'S- terminated {addr} [BAD HELLO]')

            elif read:
                print(f'S- terminated {addr} [BAD HELLO]')

            else:
                print(f'S- closed {addr} [CLIENT ABORT]')

        except BrokenPipeError:
            print(f'S- closed {addr} [BROKEN PIPE]')

        except ConnectionResetError:
            print(f'S- closed {addr} [CONNECTION RESET]')

        except socket.timeout:
            print(f'S- terminated {addr} [HELLO TIMEOUT]')

        #finally:
        #    if sock in self.peers_by_sock:
        #        self.sign_off(self.peers_by_sock[sock])

    '''
    def sign_on(self, sock, label, addr):
        self.sign_off(label)
        self.peers_by_sock[sock] = label
        self.labels[label] = addr

    def sign_off(self, label):
        if label not in self.labels: return
        del self.peers_by_sock.inverse[label]
        del self.labels[label]
    '''






def main(args):
    print('LANK: ', end='')

    cmd = args[0] if args else 'help'
    args = args[1:]

    if cmd in SET:
        cmd = SET[cmd]
        print(cmd[1])
        cmd[0](args)

    else:
        print(f'unknown command: {cmd}')
        print('Try `lank help` for a list of commands.')


def help(args):
    print('Listening Anchor Nodes for K')
    print()
    print('USAGE:')
    print('   lank <command>')
    print()
    print('The following commands are available:')
    for name, cmd in SET.items():
        print(f'   {name} - {cmd[1]}')


def version(args):
    from . import __version__
    print(f'Installed version is {__version__}')


def register(args):
    from .registration import Interactive

    Interactive().run()


def node(args):
    from .node.cmd import main as node_main

    node_main(args)


def peer(args):
    from .peer.cmd import main as peer_main

    peer_main(args)


SET = {
    'help': (help,
        'help for this program'),
    'version': (version,
        'version information'),
    'register': (register,
        'register a new label'),
    'node': (node,
        'node commands'),
    'peer': (peer,
        'peer commands'),
}






def main(args):
    print('LANK NODE: ', end='')

    cmd = args[0] if args else 'help'
    args = args[1:]

    if cmd in SET:
        cmd = SET[cmd]
        print(cmd[1])
        cmd[0](args)

    else:
        print(f'unknown command: {cmd}')
        print('Try `lank node help` for a list of commands.')


def help(args):
    print()
    print('USAGE:')
    print('   lank node <command>')
    print()
    print('The following commands are available:')
    for name, cmd in SET.items():
        print(f'   {name} - {cmd[1]}')


def run(args):
    from . import Master
    import lank.node.db as ldb

    try:
        print('Starting node process...')
        node = Master()
        node.run()

    finally:
        ldb.close()


def run2(args):
    from . import Master
    import lank.node.db as ldb

    try:
        print('Starting node process...')
        node = Master(42124)
        node.run()

    finally:
        ldb.close()


def test_v1(args):
    from .test import TestClient_v1
    tester = TestClient_v1()
    print('Testing...')
    tester.run()


def test_v2(args):
    from .test import TestClient_v2
    tester = TestClient_v2()
    print('Testing...')
    tester.run()


def dbinfo(args):
    from ..config import DB
    from .db import VERSION
    from os.path import getsize

    print(f'  db file: {DB}')
    print(f'  version: {VERSION}')
    print(f'     size: {getsize(DB)} (bytes)')


SET = {
    'help': (help,
        'help for node commands'),
    'run': (run,
        'run the node'),
    'run2': (run2,
        'run a second node'),
    'test': (test_v1,
        'test the node (protocol v1)'),
    'test2': (test_v2,
        'test the node (protocol v2)'),
    'dbinfo': (dbinfo,
        'database information'),
}




from .protocol import get_handler, HELLO

from gevent.pool import Pool
from gevent.server import StreamServer

from socket import timeout


DEFAULT_PORT = 42819

HELLO_SIZE = len(HELLO)
HELLO_TIMEOUT = 9 # seconds

GENERAL_TIMEOUT = 9 * 60 # seconds


class Server:
    def __init__(self, crypto, priv_key, port=DEFAULT_PORT):
        self.crypto = crypto
        self.priv_key = priv_key
        self.port = port

        self.pool = Pool()
        self.server = StreamServer(('0.0.0.0', self.port), self.handle,
            spawn=self.pool)

        self.buffer = bytearray(HELLO_SIZE)

    def run(self):
        self.server.serve_forever()

    def stop(self):
        self.server.stop()

    def handle(self, sock, addr):
        print(f' + connection from {addr}')
        sock.settimeout(HELLO_TIMEOUT)

        try:
            read = sock.recv_into(self.buffer)

            if read == HELLO_SIZE:
                if self.buffer == HELLO:
                    sock.settimeout(GENERAL_TIMEOUT)

                    try:
                        protocol = get_handler(sock, addr,
                            self.crypto, self.priv_key)

                        if protocol:
                            try:
                                protocol.server(self)
                                print(f' - finished {addr}')

                            except BrokenPipeError:
                                print(f' - closed {addr} [BROKEN PIPE]')

                            except ValueError as e:
                                print(f' - terminated {addr}' \
                                    + f' [BAD MESSAGE: {e}]')

                        else:
                            print(f' - closed {addr} [CLIENT ABORT]')

                    except ValueError as e:
                        print(f' - terminated {addr} [PROTOCOL VERSION]')

                    except timeout:
                        print(f' - terminated {addr} [GENERAL TIMEOUT]')

                else:
                    print(f' - terminated {addr} [BAD HELLO]')

            elif read:
                print(f' - terminated {addr} [BAD HELLO]')

            else:
                print(f' - closed {addr} [CLIENT ABORT]')

        except timeout:
            print(f' - terminated {addr} [HELLO TIMEOUT]')






def patch(cur):
    print('!!! WARNING -- PROCEED WITH CAUTION !!!')
    print('    This patch will delete any and all existing signed entries.')
    print('    Hit <ENTER> to continue...')
    input()

    cur.execute('DROP TABLE signed')

    cur.execute('''
        CREATE TABLE signed (
            id          INTEGER PRIMARY KEY,
            uuid        TEXT NOT NULL UNIQUE,
            label       INTEGER NOT NULL,
            name        INTEGER NOT NULL,
            key         TEXT NOT NULL,
            address     TEXT NOT NULL,
            signature   BLOB NOT NULL,
            version     INTEGER NOT NULL,
            node_uuid   TEXT NOT NULL,
            created     TIMESTAMP NOT NULL,

            FOREIGN KEY (label) REFERENCES label (id),
            FOREIGN KEY (name) REFERENCES name (id)
        )
    ''')

    cur.execute('''
        CREATE INDEX signed_label_name_key ON signed (
            label,
            name,
            key
        )
    ''')

    cur.execute('''
        CREATE INDEX signed_name_key ON signed (
            name,
            key
        )
    ''')

    cur.execute('''
        CREATE INDEX signed_created ON signed (
            created
        )
    ''')




from .base import Labeled, Identified

from uuid import uuid4


class Reservation(Labeled, Identified):
    def __init__(self, label, uuid=None):
        Labeled.__init__(self, label)
        Identified.__init__(self, uuid if uuid else uuid4())

    def _str_(self):
        return ', '.join([
            Labeled._str_(self),
            Identified._str_(self),
        ])

    def to_bytes(self, handler):
        return Labeled.to_bytes(self, handler) \
            + Identified.to_bytes(self, handler)

    @classmethod
    def recv(cls, handler):
        label = cls._label_(handler)
        if label is None: return None

        uuid = cls._uuid_(handler)
        if uuid is None: return None

        return cls(label, uuid)


class ReservationCancel(Labeled):
    def __init__(self, label, exists=False):
        Labeled.__init__(self, label)
        self.exists = exists

    def _str_(self):
        return ', '.join([
            Labeled._str_(self),
            f'exists={self.exists}'
        ])

    def to_bytes(self, handler):
        return Labeled.to_bytes(self, handler) \
            + (b'\xFF' if self.exists else b'\x00')

    @classmethod
    def recv(cls, handler):
        label = cls._label_(handler)
        if label is None: return None

        exists = handler.recv_bytes(1)
        if exists is None: return None

        return cls(label, exists==b'\xFF')


class ReservationRequired(Labeled):
    pass


class Registration(Labeled):
    VERSION_SIZE = 1 # bytes
    TIME_NONCE_SIZE_SIZE = 1 # bytes
    KEY_PAIR_SIZE_SIZE = 2 # bytes
    SIG_SIZE = 512 # bytes

    def __init__(self, label, version, time_nonce, key_pair_pem, signature):
        Labeled.__init__(self, label)
        self.version = version
        self.time_nonce = time_nonce
        self.key_pair_pem = key_pair_pem
        self.signature = signature

    def _str_(self):
        return ', '.join([
            Labeled._str_(self),
            f'version={self.version}',
            #f'time_nonce={self.time_nonce}',
        ])

    def to_bytes(self, handler):
        label = Labeled.to_bytes(self, handler)

        version = self.version
        assert version > 0 and version < 256**self.VERSION_SIZE
        version = version.to_bytes(self.VERSION_SIZE, handler.BYTE_ORDER)

        time_nonce = self.time_nonce.encode(handler.ENCODING)
        time_nonce_size = len(time_nonce)
        assert time_nonce_size > 0 \
            and time_nonce_size < 256**self.TIME_NONCE_SIZE_SIZE
        time_nonce_size = time_nonce_size.to_bytes(self.TIME_NONCE_SIZE_SIZE,
                                                   handler.BYTE_ORDER)

        key_pair = self.key_pair_pem
        key_pair_size = len(key_pair)
        assert key_pair_size > 0 \
            and key_pair_size < 256**self.KEY_PAIR_SIZE_SIZE
        key_pair_size = key_pair_size.to_bytes(self.KEY_PAIR_SIZE_SIZE,
                                               handler.BYTE_ORDER)

        sig = self.signature
        assert len(sig) == self.SIG_SIZE

        return label + version + time_nonce_size + time_nonce \
            + key_pair_size + key_pair + sig

    @classmethod
    def recv(cls, handler):
        label = cls._label_(handler)
        if label is None: return None

        version = handler.recv_bytes(cls.VERSION_SIZE)
        if version is None: return None
        version = int.from_bytes(version, handler.BYTE_ORDER)

        size = handler.recv_bytes(cls.TIME_NONCE_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)

        time_nonce = handler.recv_bytes(size)
        if time_nonce is None: return None
        time_nonce = str(time_nonce, handler.ENCODING)

        size = handler.recv_bytes(cls.KEY_PAIR_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)

        key_pair = handler.recv_bytes(size)
        if key_pair is None: return None
        key_pair = bytes(key_pair)

        sig = handler.recv_bytes(cls.SIG_SIZE)
        if sig is None: return None
        sig = bytes(sig)

        return cls(label, version, time_nonce, key_pair, sig)


class RegistrationSuccess(Identified):
    pass


class GetRegistration(Labeled):
    pass




from abc import ABC
import random
from datetime import datetime, timezone, timedelta
from uuid import UUID


class Message(ABC):
    def __str__(self):
        txt = self._str_()
        return self.__repr__() + (f'[{txt}]' if txt else '')

    def _str_(self):
        return None

    def to_bytes(self, handler):
        return None

    @classmethod
    def recv(cls, handler):
        return cls()


class Nonced(Message, ABC):
    NONCE_SIZE = 8 # bytes

    def __init__(self, nonce=None):
        if nonce is None:
            nonce = random.randrange(256**self.NONCE_SIZE)
        self.nonce = nonce

    def _str_(self):
        return f'nonce={self.nonce}'

    def to_bytes(self, handler):
        return self.nonce.to_bytes(self.NONCE_SIZE, handler.BYTE_ORDER)

    @classmethod
    def recv(cls, handler):
        nonce = cls._nonce_(handler)
        if nonce is None: return None
        return cls(nonce)

    @classmethod
    def _nonce_(cls, handler):
        nonce = handler.recv_bytes(cls.NONCE_SIZE)
        if nonce is None: return None
        return int.from_bytes(nonce, handler.BYTE_ORDER)


class Timestamped(Message, ABC):
    TIMESTAMP_SIZE = 6 # bytes
    TS_PRECISION = 100000
    # 6 bytes at 10-ns precision is good until year 2059

    def __init__(self, timestamp):
        self.timestamp = self._from_datetime_(timestamp)

    def _str_(self):
        time = self._to_datetime_(self.timestamp).isoformat()
        return f'timestamp={time}'

    def check_time_skew(self, now, max_skew):
        now = self._from_datetime_(now)
        skew = abs(now / self.TS_PRECISION - self.timestamp / self.TS_PRECISION)
        return skew <= max_skew

    def to_bytes(self, handler):
        return self._timestamp_bytes_(handler, self.timestamp)

    @classmethod
    def _timestamp_bytes_(cls, handler, ts):
        return ts.to_bytes(cls.TIMESTAMP_SIZE, handler.BYTE_ORDER)

    @classmethod
    def recv(cls, handler):
        timestamp = cls._timestamp_(handler)
        if timestamp is None: return None
        return cls(cls._to_datetime_(timestamp))

    @classmethod
    def _timestamp_(cls, handler):
        timestamp = handler.recv_bytes(cls.TIMESTAMP_SIZE)
        if timestamp is None: return None
        return int.from_bytes(timestamp, handler.BYTE_ORDER)

    @classmethod
    def _from_datetime_(cls, dt):
        return int(dt.timestamp() * cls.TS_PRECISION)

    @classmethod
    def _to_datetime_(cls, ts):
        return datetime.fromtimestamp(ts / cls.TS_PRECISION, timezone.utc)


class Identified(Message, ABC):
    UUID_SIZE = 16 # bytes

    def __init__(self, uuid):
        assert isinstance(uuid, UUID)
        self.uuid = uuid

    def _str_(self):
        return f'uuid={self.uuid}'

    def to_bytes(self, handler):
        return self._uuid_bytes_(handler, self.uuid)

    @classmethod
    def _uuid_bytes_(cls, handler, uuid):
        return uuid.bytes

    @classmethod
    def recv(cls, handler):
        uuid = cls._uuid_(handler)
        if uuid is None: return None
        return cls(uuid)

    @classmethod
    def _uuid_(cls, handler):
        uuid = handler.recv_bytes(cls.UUID_SIZE)
        if uuid is None: return None
        return UUID(bytes=bytes(uuid))


class Labeled(Message, ABC):
    LABEL_SIZE_SIZE = 1 # bytes

    def __init__(self, label):
        assert label
        self.label = label

    def _str_(self):
        return f'label={self.label}'

    def to_bytes(self, handler):
        label = self.label.encode(handler.ENCODING)
        size = len(label)
        assert size > 0 and size < 256**self.LABEL_SIZE_SIZE
        size = size.to_bytes(self.LABEL_SIZE_SIZE, handler.BYTE_ORDER)
        return size + label

    @classmethod
    def recv(cls, handler):
        label = cls._label_(handler)
        if label is None: return None
        return cls(label)

    @classmethod
    def _label_(cls, handler):
        size = handler.recv_bytes(cls.LABEL_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)

        label = handler.recv_bytes(size)
        if label is None: return None
        label = str(label, handler.ENCODING)
        return label






NAME_REGISTER = 10


def patch(cur):
    cur.execute('''
        CREATE TABLE label (
            id      INTEGER PRIMARY KEY,
            name    TEXT NOT NULL COLLATE NOCASE UNIQUE
        )
    ''')

    cur.execute('''
        CREATE TABLE name (
            id      INTEGER PRIMARY KEY,
            name    TEXT NOT NULL COLLATE NOCASE UNIQUE
        )
    ''')

    cur.executemany('''
        INSERT INTO name (
            id,
            name
        )
        VALUES (
            ?, ?
        )
    ''', [
        (NAME_REGISTER, 'registration'),
    ])

    cur.execute('''
        CREATE TABLE signed (
            id          INTEGER PRIMARY KEY,
            label       INTEGER NOT NULL,
            name        INTEGER NOT NULL,
            key         TEXT NOT NULL,
            address     TEXT NOT NULL,
            signature   BLOB NOT NULL,
            version     INTEGER NOT NULL,

            FOREIGN KEY (label) REFERENCES label (id),
            FOREIGN KEY (name) REFERENCES name (id)
        )
    ''')

    cur.execute('''
        CREATE INDEX signed_label_name_key ON signed (
            label,
            name,
            key
        )
    ''')

    cur.execute('''
        CREATE INDEX signed_name_key ON signed (
            name,
            key
        )
    ''')




from . import Message, Handler as Base
from .v1 import Ping, Pong

from bidict import bidict

from abc import ABC


class Labeled(Message, ABC):
    LABEL_SIZE_SIZE = 1 # bytes

    def __init__(self, label):
        super().__init__()
        self.label = label

    def __str__(self):
        return super().__str__() + f'[label={self.label}]'

    def send(self, handler):
        handler.sock.sendall(self.get_bytes_to_send(handler))

    def get_bytes_to_send(self, handler):
        label_bytes = self.label.encode(handler.ENCODING)
        size = len(label_bytes)
        assert size > 0 and size < 256**self.LABEL_SIZE_SIZE

        return size.to_bytes(self.LABEL_SIZE_SIZE, handler.BYTE_ORDER) \
            + label_bytes

    @classmethod
    def recv(cls, handler):
        return cls(cls.recv_label(handler))

    @classmethod
    def recv_label(cls, handler):
        size = handler.recv_bytes(cls.LABEL_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)
        assert size > 0

        label = handler.recv_bytes(size)
        if label is None: return None
        label = str(label, handler.ENCODING)

        return label


class Signed(Labeled, ABC):
    pass # FIXME


class SignOn(Signed):
    pass


class SignOff(Signed):
    pass


class LabelFind(Labeled):
    pass


class LabelNone(Labeled):
    pass


class LabelInfo(Labeled):
    HOST_SIZE_SIZE = 1 # bytes
    PORT_SIZE = 2 # bytes

    def __init__(self, label, host, port):
        super().__init__(label)
        self.host = host
        self.port = port

    def __str__(self):
        return self.__repr__() \
            + f'[label={self.label} host={self.host} port={self.port}]'

    def get_bytes_to_send(self, handler):
        label = super().get_bytes_to_send(handler)

        host_bytes = self.host.encode(handler.ENCODING)
        host_size = len(host_bytes)
        assert host_size > 0 and host_size < 256**self.HOST_SIZE_SIZE

        return label \
            + host_size.to_bytes(self.HOST_SIZE_SIZE, handler.BYTE_ORDER) \
            + host_bytes \
            + self.port.to_bytes(self.PORT_SIZE, handler.BYTE_ORDER)

    @classmethod
    def recv(cls, handler):
        label = cls.recv_label(handler)

        host_size = handler.recv_bytes(cls.HOST_SIZE_SIZE)
        if host_size is None: return None
        host_size = int.from_bytes(host_size, handler.BYTE_ORDER)
        assert host_size > 0

        host = handler.recv_bytes(host_size)
        if host is None: return None
        host = str(host, handler.ENCODING)

        port = handler.recv_bytes(cls.PORT_SIZE)
        if port is None: return None
        port = int.from_bytes(port, handler.BYTE_ORDER)

        return cls(label, host, port)


class Handler(Base):
    VERSION = 2

    ENCODING = 'utf-8'

    MSG_BY_ID = bidict({
        1: Ping,
        2: Pong,
        3: SignOn,
        4: SignOff,
        5: LabelFind,
        6: LabelNone,
        7: LabelInfo,
    })

    def server(self, master):
        while msg := self.recv():
            print(f'     {self.addr} -> {msg}')
            reply = None

            if isinstance(msg, Ping):
                reply = Pong(msg.nonce)

            elif isinstance(msg, SignOn):
                master.sign_on(self.sock, msg.label, self.addr)

            elif isinstance(msg, SignOff):
                master.sign_off(msg.label)

            elif isinstance(msg, LabelFind):
                if msg.label in master.labels:
                    host = master.labels[msg.label][0]
                    port = master.labels[msg.label][1]
                    reply = LabelInfo(msg.label, host, port)

                else:
                    reply = LabelNone(msg.label)

            else:
                raise ValueError(f'unhandled message: {msg}')

            if reply:
                print(f'     {self.addr} <- {reply}')
                self.send(reply)




from ..config import DB

import sqlite3
from contextlib import AbstractContextManager
from datetime import datetime
import uuid


VERSION = 3

# meta table entries
META_VERSION = 'db_version'
META_NODE_UUID = 'node_uuid'

# name table entries
NAME_REGISTER = 10


#print(f'databasing "{DB}" ...')

con = sqlite3.connect(DB, isolation_level=None,
    detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
con.row_factory = sqlite3.Row
sqlite3.register_adapter(bool, int)
sqlite3.register_converter('BOOLEAN', lambda v: bool(int(v)))
sqlite3.register_converter('TIMESTAMP',
                           lambda v: datetime.fromisoformat(str(v, 'utf-8')))
cur = con.cursor()
#cur.execute('.dbconfig defensive on')
cur.execute('PRAGMA journal_mode=WAL')
cur.execute('PRAGMA synchronous=NORMAL')
cur.execute('PRAGMA temp_store=MEMORY')
#cur.execute('PRAGMA locking_mode=EXCLUSIVE')
cur.execute('PRAGMA foreign_keys=ON')


def cur_fetch():
    for row in cur:
        return row  # returns first row only

    return None  # if there were no rows

def cur_fetchcol(name):
    row = cur_fetch()
    if not row:
        return None

    return row[name]

def cur_fetchall():
    rows = [ ]

    for row in cur:
        rows.append(row)

    return rows

def close():
    cur.execute('VACUUM')
    con.close()


class Transaction(AbstractContextManager):
    def __init__(self):
        super().__init__()

    def __enter__(self):
        cur.execute('BEGIN')
        return self

    def __exit__(self, exc_type=None, exc_value=None, traceback=None):
        if exc_type or exc_value or traceback:
            cur.execute('ROLLBACK')
        else:
            cur.execute('COMMIT')
        return None


###
### META
###

cur.execute('''
    CREATE TABLE IF NOT EXISTS meta (
        name    TEXT PRIMARY KEY,
        value   TEXT
    )
''')

cur.executemany('''
    INSERT OR IGNORE INTO meta (
        name,
        value
    )
    VALUES (
        ?, ?
    )
''', [
    (META_VERSION, VERSION),
    (META_NODE_UUID, str(uuid.uuid4()))
])

def get_meta(name):
    cur.execute('''
        SELECT value
        FROM meta
        WHERE name = ?
    ''', (
        name,
    ))

    return cur_fetchcol('value')

def set_meta(name, value):
    cur.execute('''
        UPDATE meta
        SET value = ?
        WHERE name = ?
    ''', (
        value,
        name
    ))

def _upgrade_if_needed_():
    v = int(get_meta(META_VERSION))

    if v < VERSION:
        from .patch import upgrade
        upgrade(DB, close, cur, cur_fetchall,
                set_meta, META_VERSION, v, VERSION)

_upgrade_if_needed_() # pause here and patch the db before continuing


###
### NODE
###

#cur.execute('''
#    CREATE TABLE IF NOT EXISTS node (
#        id          INTEGER PRIMARY KEY,
#        host        TEXT NOT NULL,
#        port        INTEGER NOT NULL,
#        last_uuid   TEXT,
#        last_seen   TIMESTAMP
#    )
#''')


###
### LABEL
###

cur.execute('''
    CREATE TABLE IF NOT EXISTS label (
        id      INTEGER PRIMARY KEY,
        name    TEXT NOT NULL COLLATE NOCASE UNIQUE
    )
''')

def insert_label(name):
    cur.execute('''
        INSERT INTO label (
            name
        )
        VALUES (
            ?
        )
    ''', (
        name,
    ))

    return cur.lastrowid

def get_label(label_id):
    cur.execute('''
        SELECT *
        FROM label
        WHERE id = ?
    ''', (
        label_id,
    ))

    return cur_fetch()

def get_label_by_name(name):
    cur.execute('''
        SELECT *
        FROM label
        WHERE name = ?
    ''', (
        name,
    ))

    return cur_fetch()

def list_labels():
    cur.execute('''
        SELECT *
        FROM label
        ORDER BY name
    ''')

    return cur_fetchall()


###
### NAME
###

cur.execute('''
    CREATE TABLE IF NOT EXISTS name (
        id      INTEGER PRIMARY KEY,
        name    TEXT NOT NULL COLLATE NOCASE UNIQUE
    )
''')

cur.executemany('''
    INSERT OR IGNORE INTO name (
        id,
        name
    )
    VALUES (
        ?, ?
    )
''', [
    (NAME_REGISTER, 'registration'),
])

def get_name(name_id):
    cur.execute('''
        SELECT *
        FROM name
        WHERE id = ?
    ''', (
        name_id,
    ))

    return cur_fetch()

def list_names():
    cur.execute('''
        SELECT *
        FROM name
        ORDER BY name
    ''')

    return cur_fetchall()


###
### SIGNED
###

cur.execute('''
    CREATE TABLE IF NOT EXISTS signed (
        id          INTEGER PRIMARY KEY,
        uuid        TEXT NOT NULL UNIQUE,
        label       INTEGER NOT NULL,
        name        INTEGER NOT NULL,
        key         TEXT NOT NULL,
        address     TEXT NOT NULL,
        signature   BLOB NOT NULL,
        version     INTEGER NOT NULL,
        node_uuid   TEXT NOT NULL,
        created     TIMESTAMP NOT NULL,

        FOREIGN KEY (label) REFERENCES label (id),
        FOREIGN KEY (name) REFERENCES name (id)
    )
''')

cur.execute('''
    CREATE INDEX IF NOT EXISTS signed_label_name_key ON signed (
        label,
        name,
        key
    )
''')

cur.execute('''
    CREATE INDEX IF NOT EXISTS signed_name_key ON signed (
        name,
        key
    )
''')

cur.execute('''
    CREATE INDEX IF NOT EXISTS signed_created ON signed (
        created
    )
''')

def insert_signed(uuid, label, name, key, address, signature, version,
                  node_uuid, created):
    cur.execute('''
        INSERT INTO signed (
            uuid,
            label,
            name,
            key,
            address,
            signature,
            version,
            node_uuid,
            created
        )
        VALUES (
            ?, ?, ?, ?, ?, ?, ?, ?, ?
        )
    ''', (
        uuid,
        label,
        name,
        key,
        address,
        signature,
        version,
        node_uuid,
        created
    ))

    return cur.lastrowid

def list_signed():
    cur.execute('''
        SELECT *
        FROM signed
        ORDER BY created
    ''')

    return cur_fetchall()

def get_signed_by_uuid(uuid):
    cur.execute('''
        SELECT *
        FROM signed
        WHERE uuid = ?
    ''', (
        uuid,
    ))

    return cur_fetch()

def find_signed_by_label_name(label, name, limit=None):
    sql = '''
        SELECT *
        FROM signed
        WHERE label = ? AND name = ?
        ORDER BY created DESC
    '''

    if limit:
        sql += f'LIMIT {limit}'

    cur.execute(sql, (label, name))

    return cur_fetchall()

def find_signed_since(since):
    cur.execute('''
        SELECT *
        FROM signed
        WHERE created >= ?
        ORDER BY created
    ''', (since,))

    return cur_fetchall()

def get_last_signed_created():
    cur.execute('''
        SELECT MAX(created) AS created
        FROM signed
    ''')

    return cur_fetchcol('created')



#print('db ready :-)')




from .base import Nonced, Timestamped, Identified, Labeled


class NodeOn(Nonced, Timestamped, Identified):
    def __init__(self, timestamp, uuid, synced=None, nonce=None):
        Nonced.__init__(self, nonce)
        Timestamped.__init__(self, timestamp)
        Identified.__init__(self, uuid)
        self.synced = Timestamped._from_datetime_(synced) if synced else 0

    def _str_(self):
        synced = Timestamped._to_datetime_(self.synced).isoformat() \
                 if self.synced else None

        return ', '.join([
            Nonced._str_(self),
            Timestamped._str_(self),
            Identified._str_(self),
            f'synced={synced}',
        ])

    def to_bytes(self, handler):
        return Nonced.to_bytes(self, handler) \
            + Timestamped.to_bytes(self, handler) \
            + Identified.to_bytes(self, handler) \
            + self._timestamp_bytes_(handler, self.synced)

    @classmethod
    def recv(cls, handler):
        nonce = cls._nonce_(handler)
        if nonce is None: return None

        timestamp = cls._timestamp_(handler)
        if timestamp is None: return None
        timestamp = cls._to_datetime_(timestamp)

        uuid = cls._uuid_(handler)
        if uuid is None: return None

        synced = cls._timestamp_(handler)
        if synced is None: return None
        synced = Timestamped._to_datetime_(synced)

        return cls(timestamp, uuid, synced, nonce)


class Signed(Identified, Labeled, Timestamped):
    VERSION_SIZE = 1 # bytes
    NAME_SIZE = 1 # bytes
    KEY_SIZE_SIZE = 2 # bytes
    ADDR_SIZE_SIZE = 2 # bytes
    SIG_SIZE = 512 # bytes

    def __init__(self, version, uuid, label, name_id, key, address, signature,
                 node_uuid, created):
        self.version = version
        Identified.__init__(self, uuid)
        Labeled.__init__(self, label)
        self.name = name_id
        self.key = key
        self.address = address
        self.signature = signature
        self.node_uuid = node_uuid
        Timestamped.__init__(self, created)

    def _str_(self):
        return ', '.join([
            f'version={self.version}',
            Identified._str_(self),
            Labeled._str_(self),
            f'name={self.name}',
            f'node_uuid={self.node_uuid}',
            Timestamped._str_(self),
        ])

    def to_bytes(self, handler):
        ver = self.version
        assert ver > 0 and ver < 256**self.VERSION_SIZE
        ver = ver.to_bytes(self.VERSION_SIZE, handler.BYTE_ORDER)

        uuid = Identified.to_bytes(self, handler)
        label = Labeled.to_bytes(self, handler)

        name = self.name
        assert name > 0 and name < 256**self.NAME_SIZE
        name = name.to_bytes(self.NAME_SIZE, handler.BYTE_ORDER)

        key = self.key.encode(handler.ENCODING)
        key_size = len(key)
        assert key_size > 0 and key_size < 256**self.KEY_SIZE_SIZE
        key_size = key_size.to_bytes(self.KEY_SIZE_SIZE, handler.BYTE_ORDER)

        addr = self.address.encode(handler.ENCODING)
        addr_size = len(addr)
        assert addr_size > 0 and addr_size < 256**self.ADDR_SIZE_SIZE
        addr_size = addr_size.to_bytes(self.ADDR_SIZE_SIZE, handler.BYTE_ORDER)

        sig = self.signature
        assert len(sig) == self.SIG_SIZE

        node_uuid = self._uuid_bytes_(handler, self.node_uuid)
        created = Timestamped.to_bytes(self, handler)

        return ver + uuid + label + name + key_size + key \
            + addr_size + addr + sig + node_uuid + created

    @classmethod
    def recv(cls, handler):
        ver = handler.recv_bytes(cls.VERSION_SIZE)
        if ver is None: return None
        ver = int.from_bytes(ver, handler.BYTE_ORDER)

        uuid = cls._uuid_(handler)
        if uuid is None: return None

        label = cls._label_(handler)
        if label is None: return None

        name = handler.recv_bytes(cls.NAME_SIZE)
        if name is None: return None
        name = int.from_bytes(name, handler.BYTE_ORDER)

        size = handler.recv_bytes(cls.KEY_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)
        key = handler.recv_bytes(size)
        if key is None: return None
        key = str(key, handler.ENCODING)

        size = handler.recv_bytes(cls.ADDR_SIZE_SIZE)
        if size is None: return None
        size = int.from_bytes(size, handler.BYTE_ORDER)
        addr = handler.recv_bytes(size)
        if addr is None: return None
        addr = str(addr, handler.ENCODING)

        sig = handler.recv_bytes(cls.SIG_SIZE)
        if sig is None: return None
        sig = bytes(sig)

        node_uuid = cls._uuid_(handler)
        if node_uuid is None: return None

        created = cls._timestamp_(handler)
        if created is None: return None
        created = cls._to_datetime_(created)

        return cls(ver, uuid, label, name, key, addr, sig, node_uuid, created)




from abc import ABC, abstractmethod


VERSION = 2
HELLO = b'\x04\x02\x00HOLANK\x00\x02\x04'
KEEPALIVE = 90 # seconds
MAX_TIME_SKEW = 9 # seconds

cache = { }


def get_handler(sock, addr, version=None):
    if not version:
        version = sock.recv(1)
        if not version: return None
        version = version[0]

    if version not in cache:
        try:
            exec(f'from .v{version} import Handler as Protocol_v{version}')
            exec(f'cache[{version}] = Protocol_v{version}')

        except ModuleNotFoundError as e:
            cache[version] = e

    handler = cache[version]

    if isinstance(handler, Exception):
        raise ValueError(f'protocol handler version {version}') from handler

    return handler(sock, addr)


class Handler(ABC):
    VERSION = 0 # subclass MUST change
    BUFFER_SIZE = 128 # bytes

    def __init__(self, sock, addr):
        self.sock = sock
        self.addr = addr
        self.buffer = memoryview(bytearray(self.BUFFER_SIZE))

    @abstractmethod
    def client(self, master):
        raise NotImplementedError()

    @abstractmethod
    def server(self, master):
        raise NotImplementedError()

    @abstractmethod
    def send(self, msg):
        raise NotImplementedError()

    @abstractmethod
    def recv(self):
        raise NotImplementedError()

    def hello(self):
        assert self.VERSION > 0 and self.VERSION < 256
        self.sock.sendall(HELLO + self.VERSION.to_bytes(1, 'big'))

    def c_send(self, msg):
        print(f'C    {self.addr} <- {msg}')
        self.send(msg)

    def s_send(self, msg):
        print(f'S    {self.addr} <- {msg}')
        self.send(msg)

    def c_recv(self):
        msg = self.recv()
        print(f'C    {self.addr} -> {msg}')
        return msg

    def s_recv(self):
        msg = self.recv()
        print(f'S    {self.addr} -> {msg}')
        return msg

    def recv_bytes(self, size):
        if size > self.BUFFER_SIZE:
            raise ValueError(f'request to read more than buffer allows: {size}')

        total = 0

        while total < size:
            read = self.sock.recv_into(self.buffer[total:], size - total)
            if not read: return None
            total += read

        return self.buffer[:total]




from .base import Message

from abc import ABC


class Negative(Message, ABC):
    pass


class NodeIsSelf(Negative):
    pass


class NodeAlreadyConnected(Negative):
    pass


class NodeTimeSkewed(Negative):
    pass


class NodeIsIsolated(Negative):
    pass

class NodesOnly(Negative):
    pass




from abc import ABC, abstractmethod


VERSION = 1

cache = { }


def get_handler(version=None):
    if not version:
        version = VERSION

    if version not in cache:
        try:
            exec(f'from .v{version} import Handler as Crypto_v{version}')
            exec(f'cache[{version}] = Crypto_v{version}')

        except ModuleNotFoundError as e:
            cache[version] = e

    handler = cache[version]

    if isinstance(handler, Exception):
        raise ValueError(f'crypto handler version {version}') from handler

    return handler()


class Handler(ABC):
    def __init__(self):
        pass

    '''
    @abstractmethod
    def register(self):
        raise NotImplementedError()

    @abstractmethod
    def get_private_key(self, label, password=None):
        raise NotImplementedError()

    @abstractmethod
    def get_public_key(self, label):
        raise NotImplementedError()
    '''

    @abstractmethod
    def load_private_key(self, key_pair_pem, password=None):
        raise NotImplementedError()

    @abstractmethod
    def encrypt(self, pub_key, data):
        raise NotImplementedError()

    @abstractmethod
    def decrypt(self, priv_key, data):
        raise NotImplementedError()

    @abstractmethod
    def sign(self, priv_key, data):
        raise NotImplementedError()

    @abstractmethod
    def verify(self, pub_key, data, signature):
        raise NotImplementedError()




from .base import Nonced, Labeled


class Ping(Nonced):
    pass


class Pong(Nonced):
    def __init__(self, nonce):
        assert nonce is not None
        super().__init__(nonce)


class LabelNotFound(Labeled):
    pass




from .base import Nonced


class Ping(Nonced):
    #LABEL_SIZE_SIZE = 1 # bytes

    def __init__(self, label, nonce=None):
        assert label
        self.label = label
        super().__init__(nonce)

    def _str_(self):
        return super()._str_() + ', ' + f'label={self.label}'

    def to_bytes(self, handler):
        label = self.label.encode(handler.ENCODING)
        #size = len(label)
        #assert size > 0 and size < 256**self.LABEL_SIZE_SIZE

        #return super().to_bytes(handler) \
        #    + size.to_bytes(self.LABEL_SIZE_SIZE, handler.BYTE_ORDER) \
        #    + label
        return super().to_bytes(handler) + label

    @classmethod
    def from_bytes(cls, handler, data):
        if data is None: return None
        nonce, data = cls._from_bytes_(handler, data)
        #assert len(data) > cls.LABEL_SIZE_SIZE

        #size = int.from_bytes(data[:cls.LABEL_SIZE_SIZE], handler.BYTE_ORDER)
        #assert size > 0

        #data = data[cls.LABEL_SIZE_SIZE:]
        #assert len(data) == size
        label = str(data, handler.ENCODING)

        return cls(label, nonce)


class Pong(Nonced):
    def __init__(self, nonce):
        assert nonce is not None
        super().__init__(nonce)




from .crypto import get_handler as crypto
from .node import NODES, HELLO_TIMEOUT, GENERAL_TIMEOUT, KEEPALIVE
from .node.protocol.v2 import *

from threading import Thread, Event
from queue import Queue, Full
import socket
import sys
from getpass import getpass


class Interactive:
    def __init__(self):
        self.crypto = crypto()
        print(f' - crypto handler v{self.crypto.VERSION}')

    def run(self):
        client = Client()
        client.start()

        try:
            client.ready.wait()

            if not client.go:
                print('ABORTED: unable to connect to node')
                return

            print()
            print('Ready to create/update a label with a new key pair.')
            print()

            label = input('Label: ')
            if label: label = label.strip()
            if not label or not client.go:
                print('ABORTED')
                return

            exists = client.check_label(label)
            if exists is None:
                print('ABORTED')
                return

            elif not exists:
                password = getpass('Password: ')

            else:
                print('   A label with that name already exists.')

                try:
                    priv_key = self.crypto.load_private_key(exists.key_pair_pem)

                    print('   ' \
                        + 'ERROR: The existing key is OPEN and must remain so.')
                    print('ABORTED')
                    return

                except TypeError: # (needs a password)
                    pass # this is expected

                password = getpass('Existing Password: ')
                if not password:
                    print('ABORTED')
                    return

                try:
                    priv_key = self.crypto.load_private_key(exists.key_pair_pem,
                                                            password=password)

                except ValueError as e:
                    if e.args: e = ' | '.join(e.args)
                    print(f'   ERROR: {e}')
                    print('ABORTED')
                    return

                password = getpass('New Password: ')

            if password:
                results = self.crypto.PASS_POLICY.test(password)
                if label.lower() in password.lower():
                    results.append('Contains Label Name')
                if results:
                    print('   WARNING: You have entered a WEAK PASSWORD.')
                    print('      ' \
                        + 'This makes it VERY LIKELY somebody will STEAL it.')
                    print('      The following tests FAILED:')
                    for result in results:
                        print(f'         - {result}')
                    print('      Proceed with CAUTION!')

                confirm = getpass('Confirm Password: ')
                if confirm != password:
                    print('ABORTED (passwords do not match)')
                    return

            else:
                print('   WARNING: Empty password creates an OPEN key pair.')
                print('      ' \
                    + 'This means EVERYBODY is allowed to control the label')
                print('      ' \
                    + 'FOREVER and CANNOT BE UNDONE. Proceed with CAUTION!')

                agree = input('Type AGREE to continue: ')
                if agree != 'AGREE':
                    print('ABORTED')
                    return

            print()
            print('Generating key pair...', end='')
            sys.stdout.flush()
            keys = self.crypto.make_keys(password)
            print(' [done]')

            priv_key = keys[0]
            priv_key_pem = keys[1]
            pub_key_pem = keys[2]

            print('Creating signature...', end='')
            sys.stdout.flush()
            time_nonce = self.crypto.make_time_nonce()
            msg = self.crypto.get_register_message(time_nonce)
            signature = self.crypto.sign(priv_key, msg)
            print(' [done]')

            print('Sanity check...', end='')
            sys.stdout.flush()
            assert self.crypto.verify(priv_key.public_key(), msg, signature)
            print(' [done]')

            print('Transmitting...', end='')
            sys.stdout.flush()
            if client.register_label(label, time_nonce, priv_key_pem,
                    pub_key_pem, signature, self.crypto.VERSION):
                print(' [SUCCESS]')

            else:
                print(' [FAIL]')

        finally:
            client.stop()
            client.join()


class Client(Thread):
    def __init__(self):
        super().__init__(name='registration client')
        self.go = False

        self.ready = Event()
        #self.queue = Queue(maxsize=2)
        self.input = None
        self.output = None

    def stop(self):
        #print('STOP')
        self.go = False

    def check_label(self, label):
        self.ready.clear()
        self.input = Reservation(label)
        self.ready.wait()

        if isinstance(self.output, Reservation):
            return False

        elif isinstance(self.output, ReservationCancel):
            #return False if self.output.exists else None
            # FIXME: if exists, transmit request to get the key

            if not self.output.exists:
                self._error_('LABEL RESERVATION CONFLICT')
                return None

            self.ready.clear()
            self.input = GetRegistration(label)
            self.ready.wait()

            assert isinstance(self.output, Registration)
            return self.output

        else:
            return None

    def register_label(self, label, time_nonce, priv_key_pem, pub_key_pem,
                       signature, version):
        self.ready.clear()
        self.input = Registration(label, version, time_nonce,
                                  priv_key_pem + pub_key_pem,
                                  signature)
        self.ready.wait()

        if isinstance(self.output, RegistrationSuccess):
            return True

        else:
            return False

    def run(self):
        self.go = True

        print(' - connecting to node:')
        node = None

        for addr in NODES:
            print(f'   * trying {addr}... ', end='')
            sys.stdout.flush()

            try:
                sock = socket.create_connection(addr, timeout=HELLO_TIMEOUT)
                sock.settimeout(HELLO_TIMEOUT)
                node = Handler(sock, addr)
                node.hello()
                print('[READY]')
                break

            except ConnectionRefusedError:
                print('[REFUSED]')
                node = None

            except socket.timeout:
                print('[TIMEOUT]')
                node = None

        self.ready.set()

        if not node:
            self.go = False

        while self.go:
            node.sock.settimeout(1)
            seconds = 0

            while self.go and not self.input and seconds < KEEPALIVE:
                try:
                    msg = node.recv()
                    if not msg:
                        self._error_('LOST CONNECTION')
                        return
                    else:
                        self._error_(f'UNEXPECTED RESPONSE: {msg}')
                        return
                except socket.timeout:
                    pass # this is expected

                seconds += 1

            if self.go and not self.input:
                self._handle_(node, Ping())

            if self.go and self.input:
                self._handle_(node, self.input)
                self.input = None

    def _handle_(self, node, req):
        node.sock.settimeout(GENERAL_TIMEOUT)

        try:
            node.send(req)
            resp = node.recv()

            if not resp:
                self._error_('LOST CONNECTION')
                return

            elif isinstance(resp, NodeIsIsolated):
                self._error_('NODE IS ISOLATED')
                return

            if isinstance(req, Ping):
                if not isinstance(resp, Pong):
                    self._error_(f'UNEXPECTED RESPONSE: {resp}')

                elif resp.nonce != req.nonce:
                    self._error_(f'BAD NONCE: {resp}')

            elif isinstance(req, Reservation):
                if isinstance(resp, Reservation) \
                        or isinstance(resp, ReservationCancel):
                    self.output = resp
                    self.ready.set()

                else:
                    self._error_(f'UNEXPECTED RESPONSE: {resp}')

            elif isinstance(req, Registration):
                if isinstance(resp, RegistrationSuccess):
                    self.output = resp
                    self.ready.set()

                else:
                    self._error_(f'UNEXPECTED RESPONSE: {resp}')

            elif isinstance(req, GetRegistration):
                if isinstance(resp, Registration):
                    self.output = resp
                    self.ready.set()

                else:
                    self._error_(f'UNEXPECTED RESPONSE: {resp}')

            else:
                self._error_(f'UNHANDLED REQUEST: {req}')

        except socket.timeout:
            self._error_('TIMEOUT')

        except BrokenPipeError:
            self._error_('BROKEN PIPE')

        except ConnectionResetError:
            self._error_('CONNECTION RESET')

    def _error_(self, txt):
        print()
        print(f'** ERROR ** [{txt}]')
        self.go = False
        self.ready.set()






def main(args):
    print('LANK PEER: ', end='')

    cmd = args[0] if args else 'help'
    args = args[1:]

    if cmd in SET:
        cmd = SET[cmd]
        print(cmd[1])
        cmd[0](args)

    else:
        print(f'unknown command: {cmd}')
        print('Try `lank peer help` for a list of commands.')


def help(args):
    print()
    print('USAGE:')
    print('   lank peer <command>')
    print()
    print('The following commands are available:')
    for name, cmd in SET.items():
        print(f'   {name} - {cmd[1]}')


def run(args):
    from . import Server
    from ..crypto import get_handler

    crypto = get_handler()

    label = 'anonymous'
    print(f'Loading private key for "{label}"...')
    priv_key = crypto.get_private_key(label)

    print('Starting server process...')
    server = Server(crypto, priv_key)
    print(f'   listening on port {server.port}')
    server.run()


def test(args):
    from .test import TestClient
    tester = TestClient()
    print('Testing...')
    tester.run()


SET = {
    'help': (help,
        'help for peer commands'),
    'run': (run,
        'run the peer'),
    'test': (test,
        'test the peer'),
}




from ..v1 import Handler as Base
from .ack import *
from .media import *

from bidict import bidict


class Handler(Base):
    VERSION = 2

    ENCODING = 'utf-8'
    BUFFER_SIZE = 2048 # bytes
    CRYPTO_SIG_SIZE = 512 # bytes
    CRYPTO_SIZE_SIZE = 3 # bytes

    MSG_BY_ID = bidict({
        1: Ping,
        2: Pong,
        3: Text,
    })

    def __init__(self, sock, addr, crypto, priv_key, pub_key=None):
        super().__init__(sock, addr)

        self.crypto = crypto
        self.priv_key = priv_key
        self.pub_key = pub_key
        self.label = None

    def server(self, master):
        while msg := self.recv():
            print(f'     {self.addr} -> {msg}')
            reply = None

            if isinstance(msg, Ping):
                reply = Pong(msg.nonce)

            elif isinstance(msg, Text):
                #reply = Text('got your message')
                pass

            else:
                raise ValueError(f'unhandled message: {msg}')

            if reply:
                print(f'     {self.addr} <- {reply}')
                self.send(reply)

    def send(self, msg):
        id_bytes = self.get_id_bytes(msg)
        data = msg.to_bytes(self)

        data = self.crypto.encrypt(self.pub_key,
            id_bytes + (data if data else b''))

        sig = self.crypto.sign(self.priv_key, data)
        assert len(sig) == self.CRYPTO_SIG_SIZE

        size_bytes = len(data).to_bytes(self.CRYPTO_SIZE_SIZE, self.BYTE_ORDER)

        self.sock.sendall(sig + size_bytes + data)

    def recv(self):
        sig = self.recv_bytes(self.CRYPTO_SIG_SIZE)
        if not sig: return None
        sig = bytes(sig)

        crypto_size = self.recv_bytes(self.CRYPTO_SIZE_SIZE)
        if not crypto_size: return None
        crypto_size = int.from_bytes(crypto_size, self.BYTE_ORDER)
        if not crypto_size: return None

        data = self.recv_bytes(crypto_size)
        if not data: return None
        data = bytes(data)

        verify = (data, sig)
        data = self.crypto.decrypt(self.priv_key, data)

        msg = self.get_msg_type(data[:self.ID_SIZE])
        data = data[self.ID_SIZE:]

        if not issubclass(msg, Ping):
            if not self.pub_key:
                raise ValueError(f'have signed message but no key to check')

            if not self.crypto.verify(self.pub_key, verify[0], verify[1]):
                raise ValueError(f'signature verification failed')

        msg = msg.from_bytes(self, data)

        if isinstance(msg, Ping):
            if msg.label != self.label:
                self.pub_key = self.crypto.get_public_key(msg.label)

                if not self.pub_key:
                    raise ValueError(f'no key on file for label: {msg.label}')

                self.label = msg.label
                print(f'     {self.addr} is {self.label}')

            if not self.crypto.verify(self.pub_key, verify[0], verify[1]):
                raise ValueError(f'signature verification failed')

        return msg




from . import DEFAULT_PORT
from .protocol.v2 import *
from ..crypto import get_handler

from gevent import socket

import sys
from getpass import getpass


#LABEL_SELF = 'anonymous'
LABEL_SELF = 'LANNOCC'
LABEL_OTHER = 'anonymous'


def begin(txt):
    print(f'   {txt}...', end='')
    sys.stdout.flush()

def end():
    print(' [done]')


class TestClient:
    def __init__(self):
        pass

    def run(self):
        begin('instantiating crypto handler')
        crypto = get_handler()
        end()

        begin(f'loading private key for "{LABEL_SELF}" (self)')
        try:
            priv_key = crypto.get_private_key(LABEL_SELF)

        except TypeError:
            print()
            password = getpass('      Password: ')
            priv_key = crypto.get_private_key(LABEL_SELF, password)
        end()

        begin(f'loading public key for "{LABEL_OTHER}" (other)')
        pub_key = crypto.get_public_key(LABEL_OTHER)
        end()

        addr = ('localhost', DEFAULT_PORT)
        begin(f'connecting to {addr}')
        sock = socket.create_connection(addr)
        end()

        begin('instantiating protocol handler')
        handler = Handler(sock, addr, crypto, priv_key, pub_key)
        end()

        begin(f'sending HELLO v{handler.VERSION}')
        handler.hello()
        end()

        #handler.hello()
        #handler.sock.sendall(bytes(1024))

        begin('ping-pong')
        for i in range(9):
            ping = Ping(LABEL_SELF)

            print('<', end='')
            sys.stdout.flush()
            handler.send(ping)

            msg = handler.recv()

            if isinstance(msg, Pong):
                print('>', end='')
                sys.stdout.flush()

                if msg.nonce != ping.nonce:
                    print()
                    print(f'      OOPS: sent {ping.nonce}, got {msg.nonce}')
                    break

            else:
                print()
                print(f'      OOPS: {msg}')
                break
        end()

        '''
        begin(f'switching identity to {LABEL_OTHER}')
        LABEL_SELF = LABEL_OTHER
        end()

        begin('ping-pong')
        for i in range(9):
            ping = Ping(LABEL_SELF)

            print('<', end='')
            sys.stdout.flush()
            handler.send(ping)

            msg = handler.recv()

            if isinstance(msg, Pong):
                print('>', end='')
                sys.stdout.flush()

                if msg.nonce != ping.nonce:
                    print()
                    print(f'      OOPS: sent {ping.nonce}, got {msg.nonce}')
                    break

            else:
                print()
                print(f'      OOPS: {msg}')
                break
        end()
        '''

        begin('sending a text message')
        msg = Text('Today is the beginning of the rest of your life!')
        handler.send(msg)
        end()

        '''
        begin('awaiting text reply')
        msg = handler.recv()
        if not isinstance(msg, Text):
            print()
            print(f'      OOPS: {msg}')
        end()
        '''




from abc import ABC
import random


class Message(ABC):
    def __init__(self):
        pass

    def __str__(self):
        return self.__repr__()

    def send(self, handler):
        pass

    @classmethod
    def recv(cls, handler):
        return cls()


class Nonced(Message, ABC):
    NONCE_SIZE = 4 # bytes

    def __init__(self, nonce):
        super().__init__()
        self.nonce = nonce

    def __str__(self):
        return super().__str__() + f'[nonce={self.nonce}]'

    def send(self, handler):
        handler.sock.sendall(
            self.nonce.to_bytes(self.NONCE_SIZE, handler.BYTE_ORDER))

    @classmethod
    def recv(cls, handler):
        nonce = handler.recv_bytes(cls.NONCE_SIZE)
        if nonce is None: return None
        nonce = int.from_bytes(nonce, handler.BYTE_ORDER)
        return cls(nonce)


class Ping(Nonced):
    def __init__(self, nonce=None):
        if nonce is None:
            nonce = random.randrange(256**self.NONCE_SIZE)

        super().__init__(nonce)


class Pong(Nonced):
    pass




from .base import Message


class Text(Message):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def _str_(self):
        return f'text={self.text}'

    def to_bytes(self, handler):
        return self.text.encode(handler.ENCODING)

    @classmethod
    def from_bytes(cls, handler, data):
        if data is None: return None
        return cls(str(data, handler.ENCODING))




from . import Handler as Base
#import lank.db as ldb

from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.exceptions import InvalidSignature
from password_strength import PasswordPolicy
#from password_strength.tests import (
#    Length, Uppercase, Numbers, Special, NonLetters, Strength)

from getpass import getpass
from datetime import datetime
import random
import sys


# Basic process from:
#   https://dev.to/aaronktberry/generating-encrypted-key-pairs-in-python-69b

class Handler(Base):
    VERSION = 1

    TEXT_ENCODING = 'utf-8'
    PASS_POLICY = PasswordPolicy.from_names(
        length=8, uppercase=1, numbers=1, special=1, #nonletters=1,
        strength=0.5)

    #KEY_SIZE = 15360 # too slow!
    KEY_SIZE = 4096
    KEY_PUBLIC_EXPONENT = 65537
    KEY_ENCODING = serialization.Encoding.PEM
    KEY_ENCRYPTED_FORMAT = serialization.PrivateFormat.PKCS8
    KEY_OPEN_FORMAT = serialization.PrivateFormat.TraditionalOpenSSL
    KEY_PUBLIC_FORMAT = serialization.PublicFormat.SubjectPublicKeyInfo

    SIGN_PAD = padding.PSS
    SIGN_PAD_MGF = padding.MGF1
    SIGN_PAD_MGF_ALGO = hashes.SHA256
    SIGN_PAD_SALT_LENGTH = SIGN_PAD.MAX_LENGTH
    SIGN_ALGO = hashes.SHA256

    ENCRYPT_PAD = padding.OAEP
    ENCRYPT_PAD_MGF = padding.MGF1
    ENCRYPT_PAD_MGF_ALGO = hashes.SHA256
    ENCRYPT_PAD_ALGO = hashes.SHA256
    ENCRYPT_PAD_LABEL = None

    REGISTER_MSG = 'Spread love everywhere you go. ' \
        + 'Let no one ever come to you without leaving happier.'
        # --Mother Teresa

    def make_time_nonce(self):
        return f'{datetime.now().timestamp()}|{random.randrange(2**32)}'

    def get_register_message(self, time_nonce):
        return f'{self.REGISTER_MSG}~{time_nonce}'.encode(self.TEXT_ENCODING)

    def make_keys(self, password=None):
        priv_key = rsa.generate_private_key(
            key_size=self.KEY_SIZE,
            public_exponent=self.KEY_PUBLIC_EXPONENT)

        if password:
            priv_key_pem = priv_key.private_bytes(
                encoding=self.KEY_ENCODING,
                format=self.KEY_ENCRYPTED_FORMAT,
                encryption_algorithm=serialization.BestAvailableEncryption(
                    password.encode(self.TEXT_ENCODING)))

        else:
            priv_key_pem = priv_key.private_bytes(
                encoding=self.KEY_ENCODING,
                format=self.KEY_OPEN_FORMAT,
                encryption_algorithm=serialization.NoEncryption())

        pub_key_pem = priv_key.public_key().public_bytes(
            encoding=self.KEY_ENCODING,
            format=self.KEY_PUBLIC_FORMAT)

        return (priv_key, priv_key_pem, pub_key_pem)

    '''
    def get_private_key(self, label, password=None):
        label_id = ldb.get_label_by_name(label)
        if not label_id: return None

        label_id = label_id['id']
        registration = ldb.find_signed(label_id, ldb.NAME_REGISTER, limit=1)
        assert registration is not None
        registration = registration[0]
        #FIXME: check signed version

        if password:
            password = password.encode(self.TEXT_ENCODING)

        return serialization.load_pem_private_key(
            registration['address'], password=password)

    def get_public_key(self, label):
        label_id = ldb.get_label_by_name(label)
        if not label_id: return None

        label_id = label_id['id']
        registration = ldb.find_signed(label_id, ldb.NAME_REGISTER, limit=1)
        assert registration is not None
        registration = registration[0]
        #FIXME: check signed version

        return serialization.load_pem_public_key(
            registration['address'])
    '''

    def load_private_key(self, key_pair_pem, password=None):
        if password:
            password = password.encode(self.TEXT_ENCODING)

        return serialization.load_pem_private_key(key_pair_pem,
                                                  password=password)

    def encrypt(self, pub_key, data):
        return pub_key.encrypt(
            plaintext=data,
            padding=self.ENCRYPT_PAD(
                mgf=self.ENCRYPT_PAD_MGF(algorithm=self.ENCRYPT_PAD_MGF_ALGO()),
                algorithm=self.ENCRYPT_PAD_ALGO(),
                label=self.ENCRYPT_PAD_LABEL))

    def decrypt(self, priv_key, data):
        return priv_key.decrypt(
            ciphertext=data,
            padding=self.ENCRYPT_PAD(
                mgf=self.ENCRYPT_PAD_MGF(algorithm=self.ENCRYPT_PAD_MGF_ALGO()),
                algorithm=self.ENCRYPT_PAD_ALGO(),
                label=self.ENCRYPT_PAD_LABEL))

    def sign(self, priv_key, data):
        return priv_key.sign(
            data=data,
            padding=self.SIGN_PAD(
                mgf=self.SIGN_PAD_MGF(algorithm=self.SIGN_PAD_MGF_ALGO()),
                salt_length=self.SIGN_PAD_SALT_LENGTH),
            algorithm=self.SIGN_ALGO())

    def verify(self, pub_key, data, signature):
        try:
            pub_key.verify(
                signature=signature,
                data=data,
                padding=self.SIGN_PAD(
                    mgf=self.SIGN_PAD_MGF(algorithm=self.SIGN_PAD_MGF_ALGO()),
                    salt_length=self.SIGN_PAD_SALT_LENGTH),
                algorithm=self.SIGN_ALGO())

            return True

        except InvalidSignature:
            return False




from ..v1 import Handler as Base
from .. import KEEPALIVE, MAX_TIME_SKEW
from .ack import *
from .sync import *
from .register import *
from .negative import *
import lank.node.db as ldb

from bidict import bidict
from gevent import socket

from datetime import datetime, timedelta
from uuid import UUID


class Handler(Base):
    VERSION = 2

    BUFFER_SIZE = 8192
    ENCODING = 'utf-8'
    SYNC_MARGIN = 10 * 60 # seconds

    MSG_BY_ID = bidict({
         1: Ping,
         2: Pong,
         3: NodeOn,
         4: NodeIsSelf,
         5: NodeAlreadyConnected,
         6: NodeTimeSkewed,
         7: NodeIsIsolated,
         8: Reservation,
         9: ReservationRequired,
        10: ReservationCancel,
        11: Registration,
        12: RegistrationSuccess,
        13: GetRegistration,
        14: LabelNotFound,
        15: Signed,
        16: NodesOnly,
    })

    def reservation(self, master, msg):
        assert isinstance(msg, Reservation)
        reply = None

        if not master.nodes_by_uuid:
            reply = NodeIsIsolated()

        elif msg.label in master.reservations:
            reservation = master.reservations[msg.label]
            res_uuid = reservation[0]

            if msg.uuid != res_uuid: # collision
                reply = ReservationCancel(msg.label)
                master.broadcast_nodes(reply, skip=self)
                del master.reservations[msg.label]

            else: # just an echo from another node?
                pass

        elif ldb.get_label_by_name(msg.label):
            reply = ReservationCancel(msg.label, True)

        else:
            reservation = (msg.uuid, self.sock, master.now())
            master.reservations[msg.label] = reservation
            master.broadcast_nodes(msg, skip=self)
            if not self.node_uuid: reply = msg

        return reply

    def reservation_cancel(self, master, msg):
        assert isinstance(msg, ReservationCancel)

        if msg.label in master.reservations:
            master.broadcast_nodes(msg, skip=self)
            del master.reservations[msg.label]

    def registration(self, master, msg):
        assert isinstance(msg, Registration)
        reply = None

        if not master.nodes_by_uuid:
            reply = NodeIsIsolated()

        elif msg.label not in master.reservations:
            reply = ReservationRequired(msg.label)

        else:
            reservation = master.reservations[msg.label]
            res_uuid = reservation[0]
            res_sock = reservation[1]

            if res_sock != self.sock: # imposter
                reply = ReservationRequired(msg.label)

            else:
                # FIXME: verify signature

                label_id = None
                with ldb.Transaction():
                    label_id = ldb.insert_label(msg.label)
                    time = master.now()

                    ldb.insert_signed(str(res_uuid), label_id,
                        ldb.NAME_REGISTER, msg.time_nonce,
                        str(msg.key_pair_pem, self.ENCODING),
                        msg.signature, msg.version, str(master.uuid), time)

                    master.signed_recently[res_uuid] = time
                    del master.reservations[msg.label]

                master.labels_by_id[label_id] = msg.label

                master.broadcast_nodes(Signed(
                    msg.version,
                    res_uuid,
                    msg.label,
                    ldb.NAME_REGISTER,
                    msg.time_nonce,
                    str(msg.key_pair_pem, self.ENCODING),
                    msg.signature,
                    master.uuid,
                    time
                ), skip=self)

                reply = RegistrationSuccess(res_uuid)

        return reply

    def get_registration(self, master, msg):
        assert isinstance(msg, GetRegistration)

        if msg.label not in master.labels_by_id.inverse:
            return LabelNotFound(msg.label)

        label_id = master.labels_by_id.inverse[msg.label]
        signed = ldb.find_signed_by_label_name(label_id, ldb.NAME_REGISTER,
                                               limit=1)

        assert signed
        assert len(signed)==1
        signed = signed[0]

        return Registration(
            msg.label,
            signed['version'],
            signed['key'],
            signed['address'].encode(self.ENCODING),
            signed['signature'])

    def signed(self, master, msg):
        assert isinstance(msg, Signed)

        if msg.uuid in master.signed_recently:
            return

        exists = ldb.get_signed_by_uuid(str(msg.uuid))
        if exists:
            return

        # FIXME: verify signature

        label_id = None
        with ldb.Transaction():
            try:
                label_id = master.labels_by_id.inverse[msg.label]
            except KeyError:
                label_id = ldb.insert_label(msg.label)

            ldb.insert_signed(str(msg.uuid), label_id, msg.name, msg.key,
                              msg.address, msg.signature, msg.version,
                              str(msg.node_uuid),
                              msg._to_datetime_(msg.timestamp))

        if label_id not in master.labels_by_id:
            master.labels_by_id[label_id] = msg.label

        if msg.label in master.reservations:
            del master.reservations[msg.label]

        master.signed_recently[msg.uuid] = master.now()
        master.broadcast_nodes(msg, skip=self)

    def client(self, master):
        self.node_uuid = None

        sync = ldb.get_last_signed_created()
        if sync:
            sync = datetime.fromisoformat(sync)
            sync -= timedelta(seconds=self.SYNC_MARGIN)
        msg = NodeOn(master.now(), master.uuid, sync)
        self.c_send(msg)

        reply = self.c_recv()
        if not reply: return

        if isinstance(reply, Negative):
            raise KeyError(reply.__class__.__name__)
        elif not isinstance(reply, NodeOn):
            raise ValueError('expecting NodeOn reply')
        elif reply.uuid == master.uuid:
            raise KeyError(NodeIsSelf.__name__)
        elif reply.nonce != msg.nonce:
            raise ValueError('nonce mismatch')
        elif reply.uuid in master.nodes_by_uuid:
            raise ValueError('node already connected (uuid match)')
        elif not reply.check_time_skew(master.now(), MAX_TIME_SKEW):
            raise ValueError('time skew')

        self.node_uuid = reply.uuid
        master.nodes_by_uuid[self.node_uuid] = self
        master.status()

        try:
            self.sock.settimeout(KEEPALIVE)
            ping = None

            while msg:
                try:
                    while msg := self.c_recv():
                        if isinstance(msg, Pong):
                            if not ping:
                                raise ValueError('pong without ping')
                            elif msg.nonce != ping.nonce:
                                raise ValueError('nonce mismatch')
                            ping = None

                        elif isinstance(msg, Reservation):
                            reply = self.reservation(master, msg)
                            if reply: self.c_send(reply)

                        elif isinstance(msg, ReservationCancel):
                            self.reservation_cancel(master, msg)

                        elif isinstance(msg, GetRegistration):
                            reply = self.get_registration(master, msg)
                            if reply: self.c_send(reply)

                        elif isinstance(msg, Signed):
                            self.signed(master, msg)

                        else:
                            raise ValueError('unhandled message')

                except socket.timeout as e:
                    if ping: raise e
                    ping = Ping()
                    self.c_send(ping)

        finally:
            del master.nodes_by_uuid[self.node_uuid]
            master.status()

    def server(self, master):
        self.node_uuid = None

        try:
            while msg := self.s_recv():
                reply = None

                if isinstance(msg, Ping):
                    reply = Pong(msg.nonce)

                elif isinstance(msg, NodeOn):
                    if msg.uuid == master.uuid:
                        reply = NodeIsSelf()

                    elif msg.uuid in master.nodes_by_uuid:
                        reply = NodeAlreadyConnected()

                    elif not msg.check_time_skew(master.now(), MAX_TIME_SKEW):
                        reply = NodeTimeSkewed()

                    else:
                        self.node_uuid = msg.uuid

                        sync = ldb.get_last_signed_created()
                        if sync:
                            sync = datetime.fromisoformat(sync)
                            sync -= timedelta(seconds=self.SYNC_MARGIN)

                        self.s_send(NodeOn(master.now(), master.uuid, sync,
                            msg.nonce))

                        msg_sync = msg._to_datetime_(msg.synced) \
                                   if msg.synced else None

                        if sync and (not msg_sync or msg_sync < sync):
                            if not msg_sync:
                                signed_list = ldb.list_signed()
                            else:
                                signed_list = ldb.find_signed_since(msg_sync)

                            #seq = 1
                            for signed in signed_list:
                                #print(f'XXX {signed["created"]}')
                                self.s_send(Signed(
                                    signed['version'],
                                    UUID(signed['uuid']),
                                    master.labels_by_id[signed['label']],
                                    signed['name'],
                                    signed['key'],
                                    signed['address'],
                                    signed['signature'],
                                    UUID(signed['node_uuid']),
                                    signed['created']
                                ))
                                #seq += 1

                        sync = ldb.get_last_signed_created()
                        if sync:
                            sync = datetime.fromisoformat(sync)
                            sync -= timedelta(seconds=self.SYNC_MARGIN)

                        #reply = NodeOn(master.now(), master.uuid, sync,
                        #    msg.nonce)

                        master.nodes_by_uuid[self.node_uuid] = self
                        master.status()

                elif isinstance(msg, Reservation):
                    reply = self.reservation(master, msg)

                elif isinstance(msg, ReservationCancel):
                    self.reservation_cancel(master, msg)

                elif isinstance(msg, Registration):
                    reply = self.registration(master, msg)

                elif isinstance(msg, GetRegistration):
                    reply = self.get_registration(master, msg)

                elif isinstance(msg, Signed):
                    if not self.node_uuid:
                        reply = NodesOnly()

                    else:
                        self.signed(master, msg)

                #elif isinstance(msg, SignOn):
                #    master.sign_on(self.sock, msg.label, self.addr)

                #elif isinstance(msg, SignOff):
                #    master.sign_off(msg.label)

                #elif isinstance(msg, LabelFind):
                #    if msg.label in master.labels:
                #        host = master.labels[msg.label][0]
                #        port = master.labels[msg.label][1]
                #        reply = LabelInfo(msg.label, host, port)

                #    else:
                #        reply = LabelNone(msg.label)

                else:
                    raise ValueError('unhandled message')

                if reply:
                    self.s_send(reply)

                    if isinstance(reply, Negative):
                        break

        finally:
            if self.node_uuid:
                del master.nodes_by_uuid[self.node_uuid]
                master.status()

    def send(self, msg):
        id_bytes = self.get_id_bytes(msg)
        data = msg.to_bytes(self)
        self.sock.sendall(id_bytes + (data if data else b''))




import sys


def upgrade(dbfile, close, cur, fetchall,
            set_meta, meta_version, from_ver, to_ver):
    print()
    print()
    print('#################################')
    print('    DATABASE UPGRADE REQUIRED')
    print('---------------------------------')
    print(f'  database file: {dbfile}')
    print(f'   your version: {from_ver}')
    print(f' needed version: {to_ver}')
    print('---------------------------------')
    print('The database must be patched')
    print('before it can be used by this')
    print('release.')
    print()
    print('! WARNING: It is recommended that')
    print('you make a backup copy of the')
    print('database file before continuing!!')
    print()
    print('Patches will apply incrementally.')
    print('---------------------------------')
    print('      USE <CTRL>-C TO EXIT')
    print('#################################')

    v = from_ver
    while v < to_ver:
        v += 1
        try:
            apply_patch(cur, fetchall, v)
            set_meta(meta_version, v)
            print('SUCCESS')

        except KeyboardInterrupt:
            print('ANCELLED BY USER') # not a typo (piggy back off ^C)
            close()
            sys.exit(1)

        except Exception as e:
            print('FAILED')
            raise RuntimeError(f'Failed to apply version {v} patch', e)

def apply_patch(cur, fetchall, version):
    print()
    print(f'Hit <ENTER> to apply version {version}...')
    input()

    exec(f'from .v{version} import patch as patch_v{version}')

    cur.execute('PRAGMA foreign_keys=OFF')
    cur.execute('BEGIN')

    exec(f'patch_v{version}(cur)')

    cur.execute('PRAGMA foreign_key_check')
    fails = fetchall()
    if fails:
        print('!!! FOREIGN KEY FAILURE !!!')
        for fail in fails:
            print(f'   table={fail[0]}, id={fail[1]}, column={fail[2]}')
        raise RuntimeError('one or more foreign key constraints failed')

    cur.execute('COMMIT')
    cur.execute('PRAGMA foreign_keys=ON')
    cur.execute('VACUUM')




from . import DEFAULT_PORT

from gevent import socket

import sys


def begin(txt):
    print(f'   {txt}...', end='')
    sys.stdout.flush()

def end():
    print(' [done]')


class TestClient_v1:
    def run(self):
        from .protocol.v1 import Handler, Ping, Pong

        addr = ('localhost', DEFAULT_PORT)

        begin(f'connecting to {addr}')
        sock = socket.create_connection(addr)
        end()

        begin('instantiating protocol handler')
        handler = Handler(sock, addr)
        end()

        begin(f'sending HELLO v{handler.VERSION}')
        handler.hello()
        end()

        begin('ping-pong')
        #for i in range(999):
        while True:
            ping = Ping()

            print('<', end='')
            sys.stdout.flush()
            handler.send(ping)

            msg = handler.recv()
            print('>', end='')
            sys.stdout.flush()

            if isinstance(msg, Pong):
                if msg.nonce != ping.nonce:
                    print()
                    print(f'      NONCE: sent {ping.nonce}, got {msg.nonce}')
                    break

            else:
                print()
                print(f'      {msg}')
                break
        end()


class TestClient_v2:
    TEST_LABEL = 'just_a_test'

    def run(self):
        from .protocol.v2 import Handler, Ping, Pong

        addr = ('localhost', DEFAULT_PORT)

        begin(f'connecting to {addr}')
        sock = socket.create_connection(addr)
        end()

        begin('instantiating protocol handler')
        handler = Handler(sock, addr)
        end()

        begin(f'sending HELLO v{handler.VERSION}')
        handler.hello()
        end()

        begin('ping-pong')
        #for i in range(999):
        while True:
            ping = Ping()

            print('<', end='')
            sys.stdout.flush()
            handler.send(ping)

            msg = handler.recv()
            print('>', end='')
            sys.stdout.flush()

            if isinstance(msg, Pong):
                if msg.nonce != ping.nonce:
                    print()
                    print(f'      NONCE: sent {ping.nonce}, got {msg.nonce}')
                    break

            else:
                print()
                print(f'      {msg}')
                break
        end()

        '''
        begin('label find #1')
        find = LabelFind(self.TEST_LABEL)
        handler.send(find)
        msg = handler.recv()
        print()
        print(f'      {msg}', end=' ')
        end()

        begin('sign on')
        sign_on = SignOn(self.TEST_LABEL)
        handler.send(sign_on)
        end()

        begin('label find #2')
        handler.send(find)
        msg = handler.recv()
        print()
        print(f'      {msg}', end=' ')
        end()

        begin('sign off')
        sign_off = SignOff(self.TEST_LABEL)
        handler.send(sign_off)
        end()

        begin('label find #3')
        handler.send(find)
        msg = handler.recv()
        print()
        print(f'      {msg}', end=' ')
        end()
        '''




from abc import ABC
import random


class Message(ABC):
    def __str__(self):
        txt = self._str_()
        return self.__repr__() + (f'[{txt}]' if txt else '')

    def _str_(self):
        return None

    def to_bytes(self, handler):
        return None

    @classmethod
    def from_bytes(cls, handler, data):
        return cls()


class Nonced(Message, ABC):
    NONCE_SIZE = 8 # bytes

    def __init__(self, nonce=None):
        super().__init__()
        if nonce is None:
            nonce = random.randrange(256**self.NONCE_SIZE)
        self.nonce = nonce

    def _str_(self):
        return f'nonce={self.nonce}'

    def to_bytes(self, handler):
        return self.nonce.to_bytes(self.NONCE_SIZE, handler.BYTE_ORDER)

    @classmethod
    def from_bytes(cls, handler, data):
        return cls(cls._from_bytes_(handler, data)[0])

    @classmethod
    def _from_bytes_(cls, handler, data):
        if data is None: return None
        assert len(data) >= cls.NONCE_SIZE
        return (int.from_bytes(data[:cls.NONCE_SIZE], handler.BYTE_ORDER),
                data[cls.NONCE_SIZE:])

